package dev.fyloz.trial.colorrecipesexplorer.core.services;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import dev.fyloz.trial.colorrecipesexplorer.core.Preferences;
import dev.fyloz.trial.colorrecipesexplorer.core.exception.model.EntityAlreadyExistsException;
import dev.fyloz.trial.colorrecipesexplorer.core.exception.model.EntityNotFoundException;
import dev.fyloz.trial.colorrecipesexplorer.core.exception.model.ModelException;
import dev.fyloz.trial.colorrecipesexplorer.core.exception.model.NullIdentifierException;
import dev.fyloz.trial.colorrecipesexplorer.core.model.IModel;
import org.slf4j.Logger;
import org.springframework.data.jpa.repository.JpaRepository;
import org.springframework.lang.NonNull;
import org.springframework.transaction.annotation.Transactional;

import javax.validation.constraints.NotNull;
import java.util.Collection;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

public abstract class AbstractService<E extends IModel, R extends JpaRepository<E, Long>> implements IGenericService<E> {

    protected Logger logger = Preferences.logger;
    protected R dao;
    protected Class<E> type;

    public AbstractService(Class<E> type) {
        this.type = type;
    }

    @Override
    public boolean exists(E entity) {
        return entity != null && entity.getId() != null && existsById(entity.getId());
    }

    @Override
    public boolean existsById(Long id) {
        return dao.existsById(id);
    }

    @Override
    public E getById(Long id) {
        Optional<E> found = dao.findById(id);
        if (found.isEmpty()) throw new EntityNotFoundException(type, EntityNotFoundException.IdentifierType.ID, id);

        return found.get();
    }

    @Override
    public List<E> getAll() {
        return dao.findAll();
    }

    @Override
    public E save(@NotNull E entity) {
        if (entity.getId() != null && existsById(entity.getId()))
            throw new EntityAlreadyExistsException(type, ModelException.IdentifierType.ID, entity.getId());

        return dao.save(entity);
    }

    @Override
    @Transactional
    public Collection<E> saveAll(@NotNull Collection<E> entities) {
        return entities
                .stream()
                .map(this::save)
                .collect(Collectors.toList());
    }

    @Override
    public E update(@NonNull E entity) {
        if (entity.getId() == null) throw new NullIdentifierException(type);
        if (!existsById(entity.getId()))
            throw new EntityNotFoundException(type, ModelException.IdentifierType.ID, entity.getId());

        return dao.save(entity);
    }

    @Override
    public void delete(@NonNull E entity) {
        dao.delete(entity);
    }

    @Override
    public void deleteById(Long id) {
        delete(getById(id));
    }

    @Override
    public void deleteAll(Collection<E> entities) {
        dao.deleteAll(entities);
    }

    /**
     * Vérifie si une entité est valide pour la création dans la base de données.
     *
     * @param entity L'entité à vérifier.
     * @return Si l'entité est valide pour la création.
     */
    @Deprecated(since = "1.3.0", forRemoval = true)
    public boolean isValidForCreation(E entity) {
        return entity != null && !existsById(entity.getId());
    }

    /**
     * Vérifie si une entité est valide pour la création dans la base de données.
     *
     * @param entity L'entité à vérifier.
     * @return Si l'entité est valide pour l'édition.
     */
    @Deprecated(since = "1.3.0", forRemoval = true)
    public boolean isValidForUpdate(@NonNull E entity) {
        return entity.getId() != null && existsById(entity.getId());
    }

    /**
     * Transforme un objet en Json.
     *
     * @param obj L'objet à transformer
     * @return L'objet sous forme de Json.
     */
    public String asJson(Object obj) {
        ObjectMapper objectMapper = new ObjectMapper();
        try {
            return objectMapper.writeValueAsString(obj);
        } catch (JsonProcessingException e) {
            Preferences.logger.error("Une erreur est survenue lors de la transformation d'un objet en Json", e);
            return null;
        }
    }
}
