package dev.fyloz.trial.colorrecipesexplorer.core.services.model;

import dev.fyloz.trial.colorrecipesexplorer.core.exception.model.EntityAlreadyExistsException;
import dev.fyloz.trial.colorrecipesexplorer.core.exception.model.EntityNotFoundException;
import dev.fyloz.trial.colorrecipesexplorer.core.exception.model.ModelException;
import dev.fyloz.trial.colorrecipesexplorer.core.model.Material;
import dev.fyloz.trial.colorrecipesexplorer.core.model.MaterialType;
import dev.fyloz.trial.colorrecipesexplorer.core.services.AbstractService;
import dev.fyloz.trial.colorrecipesexplorer.core.services.files.SimdutService;
import dev.fyloz.trial.colorrecipesexplorer.dao.MaterialDao;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.web.multipart.MultipartFile;

import javax.validation.constraints.NotNull;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

@Service
public class MaterialService extends AbstractService<Material, MaterialDao> {

    private MixQuantityService mixQuantityService;
    private SimdutService simdutService;

    public MaterialService() {
        super(Material.class);
    }

    @Autowired
    public void setMaterialDao(MaterialDao materialDao) {
        this.dao = materialDao;
    }

    @Autowired
    public void setMixQuantityService(MixQuantityService mixQuantityService) {
        this.mixQuantityService = mixQuantityService;
    }

    @Autowired
    public void setSimdutService(SimdutService simdutService) {
        this.simdutService = simdutService;
    }

    /**
     * Vérifie si un produit correspondant à un nom existe.
     *
     * @param name Le nom du produit
     * @return Si un produit correspondant au nom existe
     */
    public boolean existsByName(String name) {
        return dao.existsByName(name);
    }

    /**
     * Vérifie si des produits sont d'un type de produit
     *
     * @param materialType Le type de produit
     * @return Si des produits sont du type de produit
     */
    public boolean existsByMaterialType(MaterialType materialType) {
        return dao.existsByMaterialType(materialType);
    }

    /**
     * Vérifie si un produit est lié à un ou plusieurs mélanges.
     *
     * @param material Le produit à vérifier.
     * @return Si le produit est lié à d'autres mélanges.
     */
    public boolean isLinkedToMixes(Material material) {
        return mixQuantityService.existsByMaterial(material);
    }

    /**
     * Récupère tous les produits qui ne sont pas des types de mélange.
     *
     * @return Tous les produits qui ne sont pas des types de mélange
     */
    public List<Material> getAllNotMixType() {
        return getAll()
                .stream()
                .filter(m -> !m.isMixType())
                .collect(Collectors.toList());
    }

    /**
     * Récupère le produit correspondant à un nom.
     *
     * @param name Le nom du produit à récupérer
     * @return Le produit correspondant au nom
     */
    public Material getByName(String name) {
        Optional<Material> found = dao.findByName(name);
        if (found.isEmpty()) throw new EntityNotFoundException(type, ModelException.IdentifierType.NAME, name);

        return found.get();
    }

    public Material save(@NotNull Material material, MultipartFile file) {
        simdutService.write(material, file);

        return save(material);
    }

    @Override
    public Material update(Material material) {
        Optional<Material> materialByName = dao.findByName(material.getName());
        if (materialByName.isPresent() && !material.getId().equals(materialByName.get().getId()))
            throw new EntityAlreadyExistsException(type, ModelException.IdentifierType.NAME, material.getName());

        return super.update(material);
    }

    public Material update(Material material, MultipartFile simdut) {
        simdutService.update(simdut, material);

        return update(material);
    }

    @Override
    public void delete(Material material) {
        simdutService.delete(material);

        super.delete(material);
    }

    @Deprecated(since = "1.3.0", forRemoval = true)
    public void deleteIfNotLinked(Material material) {
        if (!isLinkedToMixes(material)) {
            delete(material);
        }
    }

}
