package dev.fyloz.trial.colorrecipesexplorer.services;

import dev.fyloz.trial.colorrecipesexplorer.core.exception.model.EntityAlreadyExistsException;
import dev.fyloz.trial.colorrecipesexplorer.core.exception.model.EntityNotFoundException;
import dev.fyloz.trial.colorrecipesexplorer.core.exception.model.ModelException;
import dev.fyloz.trial.colorrecipesexplorer.core.exception.model.NullIdentifierException;
import dev.fyloz.trial.colorrecipesexplorer.core.model.IModel;
import dev.fyloz.trial.colorrecipesexplorer.core.services.IGenericService;
import org.junit.jupiter.api.Test;
import org.springframework.data.jpa.repository.JpaRepository;

import java.util.*;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyLong;
import static org.mockito.Mockito.*;

public abstract class AbstractServiceTest<E extends IModel, S extends IGenericService<E>, R extends JpaRepository<E, Long>> {

    private S service;
    private R repository;

    public void init(S service, R repository) {
        this.service = service;
        this.repository = repository;

        E entity = getEntity();

        when(repository.existsById(anyLong())).thenReturn(false);
        when(repository.existsById(entity.getId())).thenReturn(true);

        when(repository.findById(anyLong())).thenReturn(Optional.empty());
        when(repository.findById(entity.getId())).thenReturn(Optional.of(entity));

        when(repository.save(any())).thenAnswer(i -> i.getArgument(0));
    }



    @Test
    public void whenExistsById_thenReturnTrue() {
        // given
        E entity = getEntity();

        // when
        boolean found = service.existsById(entity.getId());

        // then
        assertTrue(found);
    }

    @Test
    public void whenExistsById_thenReturnFalse() {
        // given
        E entity = getEntityDifferentId();

        // when
        boolean found = service.existsById(entity.getId());

        // then
        assertFalse(found);
    }

    @Test
    public void whenGetById_thenReturnEntity() {
        // given
        E entity = getEntity();

        // when
        E found = service.getById(entity.getId());

        // then
        assertEquals(entity, found);
    }

    @Test
    public void whenGetById_thenThrowEntityNotFoundException() {
        // given
        E entity = getEntityDifferentId();

        // when
        EntityNotFoundException thrown = assertThrows(EntityNotFoundException.class, () -> service.getById(entity.getId()));
        assertEquals(thrown.getIdentifierType(), ModelException.IdentifierType.ID);
    }

    @Test
    public void whenGetAll_thenReturnEntityList() {
        // given
        E entity = getEntity();
        E anotherEntity = getEntityDifferentId();
        List<E> entities = Arrays.asList(entity, anotherEntity);

        when(repository.findAll()).thenReturn(entities);

        // when
        List<E> found = service.getAll();

        // then
        assertFalse(found.isEmpty());
        assertTrue(found.containsAll(entities));
    }

    @Test
    public void whenGetAll_thenReturnEmptyList() {
        // given
        when(repository.findAll()).thenReturn(new ArrayList<>());

        // when
        List<E> found = service.getAll();

        // then
        assertTrue(found.isEmpty());
    }

    @Test
    public void whenSave_thenReturnEntity() {
        // given
        E entity = getEntityDifferentId();

        // when
        E found = service.save(entity);

        // then
        assertEquals(entity, found);
    }

    @Test
    public void whenSave_thenThrowEntityAlreadyExistsException() {
        // given
        E entity = getEntity();

        // then
        EntityAlreadyExistsException thrown = assertThrows(EntityAlreadyExistsException.class, () -> service.save(entity));
        assertEquals(thrown.getIdentifierType(), ModelException.IdentifierType.ID);
    }

    @Test
    public void whenSaveAll_thenReturnEntityCollection() {
        // given
        E entity = getEntity();
        E anotherEntity = getEntityDifferentId();
        List<E> entities = Arrays.asList(entity, anotherEntity);

        doAnswer(i -> i.getArgument(0)).when(service).saveAll(any());

        // when
        Collection<E> found = service.saveAll(entities);

        // then
        assertFalse(found.isEmpty());
        assertTrue(found.containsAll(entities));
    }

    @Test
    public void whenSaveAll_thenReturnEmptyCollection() {
        // given
        List<E> entities = new ArrayList<>();

        doAnswer(i -> i.getArgument(0)).when(service).saveAll(any());

        // when
        Collection<E> found = service.saveAll(entities);

        // then
        assertTrue(found.isEmpty());
    }

    @Test
    public void whenUpdate_thenReturnEntity() {
        // given
        E entity = getEntity();

        // when
        E found = service.update(entity);

        // then
        assertEquals(entity, found);
    }

    @Test
    public void whenUpdate_thenThrowNullIdentifierException() {
        // given
        E entity = getEntityNullId();

        // then
        assertThrows(NullIdentifierException.class, () -> service.update(entity));
    }

    @Test
    public void whenUpdate_thenThrowEntityNotFoundException() {
        // given
        E entity = getEntityDifferentId();

        // when
        EntityNotFoundException thrown = assertThrows(EntityNotFoundException.class, () -> service.update(entity));

        // then
        assertEquals(thrown.getIdentifierType(), ModelException.IdentifierType.ID);
    }

    @Test
    public void whenDelete_thenDaoIsCalled() {
        // given
        E entity = getEntity();

        // when
        service.delete(entity);

        // then
        verify(repository).delete(entity);
    }

    @Test
    public void whenDeleteById_thenDeleteIsCalled() {
        // given
        E entity = getEntity();

        // when
        service.deleteById(entity.getId());

        // then
        verify(service).delete(entity);
    }

    @Test
    public void whenDeleteAll_thenDaoIsCalled() {
        // given
        E entity = getEntity();
        E anotherEntity = getEntityDifferentId();
        Collection<E> entities = Arrays.asList(entity, anotherEntity);

        // when
        service.deleteAll(entities);

        // then
        verify(repository).deleteAll(entities);
    }



    protected abstract E getEntity();

    protected abstract E getEntityDifferentId();

    protected abstract E getEntityNullId();

}
